import path from 'path';
import isGlob from 'is-glob';
import mm from 'micromatch';
import fs from 'fs';

/**
 * import文のrootからのパスを求める
 */
var resolveImportPath = (importPath, relativeFilePath) => {
    // { [importAlias: string]: OriginalPath }
    const importAliasMap = {};
    // Load tsconfig option
    // MEMO: tscとか使って簡単に読める方法がありそう
    try {
        const tsConfigFile = fs.readFileSync(path.join(process.cwd(), '/tsconfig.json'), 'utf-8');
        // Exists ts config
        const tsConfig = JSON.parse(tsConfigFile);
        const compilerOptions = tsConfig.compilerOptions;
        if (compilerOptions && compilerOptions.paths) {
            const paths = compilerOptions.paths;
            Object.keys(paths).forEach((key) => {
                // FIXME: このlint ruleではimport先が存在するかチェックしておらず、複数のパスから正しい方を選択できないため[0]固定
                importAliasMap[key] = compilerOptions.baseUrl ? path.join(compilerOptions.baseUrl, paths[key][0]) : paths[key][0];
            });
        }
    }
    catch (error) {
        // DO NOTHING
    }
    if (relativeFilePath && (importPath.startsWith('./') || importPath.startsWith('../'))) {
        importPath = path.join(path.dirname(relativeFilePath), importPath);
    }
    return Object.keys(importAliasMap).reduce((resolvedImportPath, key) => {
        // FIXME: use glob module instead of replace('*')
        return resolvedImportPath.replace(key.replace('*', ''), importAliasMap[key].replace('*', ''));
    }, importPath);
};

/**
 * pathのmatcher。
 * eslintrcで設定できる値は以下のケースを扱う
 * - globパターン指定
 * - globパターン以外の場合 => 前方部分一致
 */
const isMatch = (str, pattern) => isGlob(pattern) ? mm.isMatch(str, pattern) : str.startsWith(pattern);
var strictDependencies = {
    defaultOptions: [[]],
    meta: {
        type: 'suggestion',
        messages: {
            importNotAllowed: 'import {{importPath}} is not allowed from {{relativeFilePath}}.',
        },
        schema: [
            {
                type: 'array',
                items: [
                    {
                        type: 'object',
                        properties: {
                            module: {
                                type: 'string',
                            },
                            allowReferenceFrom: {
                                type: 'array',
                                items: [
                                    {
                                        type: 'string',
                                    },
                                ],
                            },
                            allowSameModule: {
                                type: 'boolean',
                            },
                            allowTypeImport: {
                                type: 'boolean',
                            },
                        },
                    },
                ],
            },
            {
                type: 'object',
                properties: {
                    resolveRelativeImport: {
                        type: 'boolean',
                    },
                },
            },
        ],
    },
    create: (context) => {
        const dependencies = context.options[0];
        // eslint-disable-next-line no-magic-numbers
        const options = (context.options.length > 1 ? context.options[1] : {});
        const resolveRelativeImport = options.resolveRelativeImport;
        const allowTypeImport = options.allowTypeImport;
        function checkImport(node) {
            if (allowTypeImport && node.importKind === 'type') {
                return;
            }
            const fileFullPath = context.getFilename();
            const relativeFilePath = path.relative(process.cwd(), fileFullPath);
            const importPath = resolveImportPath(node.source.value, resolveRelativeImport ? relativeFilePath : undefined);
            dependencies
                .filter((dependency) => isMatch(importPath, dependency.module))
                .forEach((dependency) => {
                const isAllowed = 
                // 参照元が許可されている
                dependency.allowReferenceFrom.some((allowPath) => isMatch(relativeFilePath, allowPath)) || // または同一モジュール間の参照が許可されている場合
                    (dependency.allowSameModule && isMatch(relativeFilePath, dependency.module)) ||
                    (dependency.allowTypeImport && node.importKind === 'type');
                if (!isAllowed) {
                    context.report({
                        node,
                        messageId: 'importNotAllowed',
                        data: { importPath, relativeFilePath },
                    });
                }
            });
        }
        return {
            ImportDeclaration: checkImport,
        };
    },
};

/* eslint-disable */
module.exports = {
    rules: {
        'strict-dependencies': strictDependencies,
    },
};
